/**
 *  Demonstrate list searching and sorting algorithms. */
public class ListUtil {
  
  /** For conditional compilation of debug information. */
  public static final boolean DEBUG = false;
    
  /** For debug output, set this flag to true. */
  public static boolean BINARY_SEARCH_CHATTY;

  /** Search an array of comparable objects which is sorted
    *  in increasing order for a given item.  If the item is
    *  found in the array, return the index of the item.  
    *  Requires: list and item are not null.     
    *  @param list the array of comparable items.
    *  @param item the target to be searched for. 
    *  @returns index >= 0 if the item was found to match
    *           the one at list[index], or index < 0 if the 
    *           item was not found, but belongs at location
    *           -(index+1) in the list[] array. */
  public static int binarySearch(Comparable[] list, 
                                 Comparable item) {
    return binarySearch(list, list.length, item);
  }
  
  /**  Helper method.  Search an initial sublist of an
    *  array of comparable objects.  The sublist is sorted
    *  in increasing order for a given item.  If the item is
    *  found in the sublist, return the index of the item.  
    *  Requires: list and item are not null. 
    *  Requires: len <= list.length
    *  @param list the array of comparable items.
    *  @param len the length of the sublist.
    *  @param item the target to be searched for. 
    *  @returns index >= 0 if the item was found to match
    *           the one at list[index], or index < 0 if the 
    *           item was not found, but belongs at location
    *           -(index+1) in the list[] array. */
  public static int binarySearch(Comparable[] list, 
                                 int len,
                                 Comparable item) {
    int first = 0;
    int last = len;
    if (last == 0) {
      // Empty list
      return -1;  // -(first + 1)
    }
    int count = 0; // Keep track of the number of comparisons
    int res;
    // Loop invariant:  If the item is in the list then
    // it must have index j s.t. first <= j < last
    while (last > first + 1) {
      int m = first + (last - first)/2;
      if (DEBUG) 
        assert ( first < m && m < last ): "Middle equals first or last";
      res = list[m].compareTo(item);
      if (DEBUG) 
        count++;
      if (res <= 0)
        first = m;
      else
        last = m;
    }
    res = list[first].compareTo(item);
    if (DEBUG) {
      count++;
      if (BINARY_SEARCH_CHATTY) {
        System.out.println("Used " + count +
                           " comparisons in a list of length " +
                           list.length);
        int logFormula = 
          ((int) Math.ceil(Math.log(list.length)/Math.log(2.0))) + 1;
        System.out.println("ceil(log2(" + list.length +
                           ")) + 1 = " + logFormula);
      }
    }
    if (res == 0)
      return first;
    if (res > 0)
      return -(first+1);  // item should go in location first
    return -(first+2);    // item should go in location first+1
  }

  public static void selectionSort(Comparable list[]) {
    int n = list.length;
    for (int k= 0; k < n-1; k++ ) {
      // Find the smallest item in the remaining n-k items.
      Comparable minVal = list[k];
      int kMin = k;
      for (int j = k+1; j < n; j++) {
        if (minVal.compareTo(list[j]) > 0) {
          minVal = list[j];
          kMin = j;
        }
      }
      // Swap the minimum value into location k
      list[kMin] = list[k];
      list[k] = minVal;
    }
  }

  public static void bubbleSort(Comparable list[]) {
    int n = list.length;
    int lastSwap = n-1;
    while (lastSwap > 0) {
      // Sweep through items 0 .. lastSwap-1, swapping
      // any out of order items.
      int jSwap = -1;
      for (int j = 0; j < lastSwap; j++) {
        Comparable bubble = list[j];
        if (bubble.compareTo(list[j+1]) > 0) {
          list[j] = list[j+1];
          list[j+1] = bubble;
          jSwap = j;
        }
      }
      lastSwap = jSwap;
    }
  }

  public static void insertionSort(Comparable list[]) {
    int n = list.length;
    for (int k = 1; k < n; k++) {
      // Use binary search to locate where
      // list[k] should go in the sorted sublist,
      // consisting of list[0..k-1].
      int index = binarySearch(list, k, list[k]);
      if (index < 0)
        index = -(index + 1);
      Comparable tmp = list[k];
      for (int j = k; j > index; j--) {
        list[j] = list[j-1];
      }
      list[index] = tmp;
    }
  }

  public static void mergeSort(Comparable list[]) {
    java.util.Arrays.sort(list);
  }
}
